@@ -29,6 +29,16 @@ class NetlinkError(Exception):
     pass
 
 
+class NetlinkSocketError(socket.error):
+    '''
+    Socket error
+    '''
+    def __init__(self, code, msg=None):
+        msg = msg or os.strerror(code)
+        super(NetlinkSocketError, self).__init__(code, msg)
+        self.code = code
+
+
 class NetlinkQueueEmpty(NetlinkError):
     '''
     Timeout reached on a message queue polling
@@ -172,13 +182,6 @@ def _repr_sockets(sockets, mode):
     return ret
 
 
-class netlink_error(socket.error):
-    def __init__(self, code, msg=None):
-        msg = msg or os.strerror(code)
-        super(netlink_error, self).__init__(code, msg)
-        self.code = code
-
-
 class marshal(object):
     '''
     Generic marshalling class
@@ -219,7 +222,7 @@ class marshal(object):
                     self.buf.seek(16)
                     code = abs(struct.unpack('i', self.buf.read(4))[0])
                     if code > 0:
-                        error = netlink_error(code)
+                        error = NetlinkSocketError(code)
 
                 self.buf.seek(offset)
                 msg_class = self.msg_map.get(msg_type, nlmsg)
@@ -487,7 +490,8 @@ class iothread(threading.Thread):
                     send = self.families[family]
                     self.rtable[sock] = send
                     rsp['cmd'] = IPRCMD_ACK
-                # TODO
+                # TODO tags: remote
+                #
                 # * subscribe requests
                 # * ...
             elif cmd['cmd'] == IPRCMD_REGISTER:
@@ -791,12 +795,16 @@ class netlink(object):
         return _repr_sockets(self._sockets, 'remote')
 
     def shutdown_servers(self, *urls):
-        self._shutdown_sockets(self.iothread.servers, 'local',
-                               self.iothread.remove_server, *urls)
+        self._shutdown_sockets([i for i in self.iothread.servers
+                                if _repr_sockets([i], 'remote') !=
+                                _repr_sockets([self.iothread.sctl], 'remote')],
+                               'local', self.iothread.remove_server, *urls)
 
     def shutdown_clients(self, *urls):
-        self._shutdown_sockets(self.iothread.clients, 'remote',
-                               self.iothread.remove_client, *urls)
+        self._shutdown_sockets([i for i in self.iothread.clients
+                                if _repr_sockets([i], 'remote') !=
+                                _repr_sockets([self.iothread.sctl], 'remote')],
+                               'remote', self.iothread.remove_client, *urls)
 
     def shutdown_sockets(self, *urls):
         self._shutdown_sockets(self._sockets, 'remote',
@@ -812,6 +820,12 @@ class netlink(object):
                 except:
                     pass
 
+    def shutdown(self):
+        self.shutdown_sockets()
+        self.shutdown_clients()
+        self.shutdown_servers()
+        self._remote_cmd(self.iothread.control, IPRCMD_STOP)
+
     def serve(self, url, key=None, cert=None, ca=None):
         if key:
             self.ssl_keys[url] = ssl_credentials(key, cert, ca)
