@@ -17,7 +17,8 @@ import uuid
 import threading
 from socket import AF_INET
 from socket import AF_INET6
-from pyroute2.netlink import netlink_error
+from pyroute2.netlink import NetlinkSocketError
+from pyroute2.netlink import NetlinkQueueEmpty
 from pyroute2.netlink.iproute import iproute
 from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
 
@@ -47,7 +48,7 @@ def get_addr_nla(msg):
     return nla
 
 
-class ipset(set):
+class LinkedSet(set):
 
     def __init__(self, *argv, **kwarg):
         set.__init__(self, *argv, **kwarg)
@@ -67,7 +68,7 @@ class ipset(set):
         set.remove(self, key)
 
     def connect(self, link):
-        assert isinstance(link, ipset)
+        assert isinstance(link, LinkedSet)
         self.links.append(link)
 
 
@@ -142,22 +143,36 @@ class rwState(object):
 
 
 class IPDBError(Exception):
-    pass
+    message = None
+
+    def __init__(self, message=None, error=None):
+        message = message or self.message
+        Exception.__init__(self, message)
+        self.cause = error
+
+
+class IPDBUnrecoverableError(IPDBError):
+    message = 'unrecoverable IPDB error, restart the instance'
 
 
 class IPDBTransactionRequired(IPDBError):
-    pass
+    message = 'begin() a transaction first'
+
+
+class IPDBModeError(IPDBError):
+    message = 'wrong transaction mode for the operation'
 
 
 def update(f):
     def decorated(self, *argv, **kwarg):
         # obtain update lock
         tid = None
+        direct = True
         with self._snapshot_lock:
             # fix the pass-through mode:
             with self._direct_state:
-                kwarg['direct'] = self._direct_state.is_set()
-                if not kwarg['direct']:
+                direct = self._direct_state.is_set()
+                if not direct:
                     # 1. begin transaction for 'direct' type
                     if self._mode == 'direct':
                         tid = self.begin()
@@ -171,18 +186,47 @@ def update(f):
                             raise IPDBTransactionRequired()
                     # 4. transactions can not require transactions :)
                     elif self._mode == 'snapshot':
-                        kwarg['direct'] = True
+                        direct = True
                     # do not support other modes
                     else:
                         raise IPDBError('transaction mode not supported')
                     # now that the transaction _is_ open
-                f(self, *argv, **kwarg)
+                f(self, direct, *argv, **kwarg)
         if tid:
             # close the transaction for 'direct' type
             self.commit(tid)
     return decorated
 
 
+class IPLinkRequest(dict):
+
+    def __init__(self, interface=None):
+        dict.__init__(self)
+        if interface:
+            self.update(interface)
+
+    def update(self, interface):
+        for key in interface:
+            if interface[key] is not None:
+                self[key] = interface[key]
+
+    def __setitem__(self, key, value):
+        if key == 'kind':
+            if 'IFLA_LINKINFO' not in self:
+                self['IFLA_LINKINFO'] = {'attrs': []}
+            nla = ['IFLA_INFO_KIND', value]
+            # FIXME: we need to replace, not add
+            self['IFLA_LINKINFO']['attrs'].append(nla)
+        elif key == 'vlan_id':
+            if 'IFLA_LINKINFO' not in self:
+                self['IFLA_LINKINFO'] = {'attrs': []}
+            nla = ['IFLA_INFO_DATA', {'attrs': [['IFLA_VLAN_ID', value]]}]
+            # FIXME: we need to replace, not add
+            self['IFLA_LINKINFO']['attrs'].append(nla)
+        else:
+            dict.__setitem__(self, key, value)
+
+
 class interface(dotkeys):
     '''
     Objects of this class represent network interface and
@@ -226,16 +270,22 @@ class interface(dotkeys):
                         'stats',
                         'stats64')
         self.last_error = None
+        self._exists = False
         self._direct_state = rwState()
         self._parent = parent
         self._tids = []
         self._transactions = {}
         self._mode = mode
-        self._slaves = dotkeys()
         self._snapshot_lock = threading.RLock()
         # local setup: direct state is required
         self._direct_state.acquire()
-        self['ipaddr'] = ipset()
+        self['ipaddr'] = LinkedSet()
+        self['ports'] = LinkedSet()
+        for i in nla_fields:
+            self[i] = None
+        self['flags'] = 0
+        for i in ('state', 'change', 'mask'):
+            del self[i]
         self._direct_state.release()
         # 8<-----------------------------------
         if dev is not None:
@@ -257,24 +307,32 @@ class interface(dotkeys):
             for key in tuple(self.keys()):
                 if key in nla_fields:
                     res[key] = self[key]
-            res['ipaddr'] = ipset(self['ipaddr'])
+            res['ipaddr'] = LinkedSet(self['ipaddr'])
+            res['ports'] = LinkedSet(self['ports'])
             if self._parent is not None and not detached:
                 self['ipaddr'].connect(res['ipaddr'])
+                self['ports'].connect(res['ports'])
             return res
 
     def __sub__(self, pif):
         '''
         '''
         res = interface(ipr=self.ip, mode='snapshot')
+        self._direct_state.acquire()
         # simple keys
         for key in self:
             if (key in nla_fields) and \
                     ((key not in pif) or (self[key] != pif[key])):
                 res[key] = self[key]
+        self._direct_state.release()
         # ip addresses
-        ipaddr = ipset(self['ipaddr'] - pif['ipaddr'])
+        ipaddr = LinkedSet(self['ipaddr'] - pif['ipaddr'])
+        # ports
+        ports = LinkedSet(self['ports'] - pif['ports'])
         if ipaddr:
             res['ipaddr'] = ipaddr
+        if ports:
+            res['ports'] = ports
 
         return res
 
@@ -282,7 +340,11 @@ class interface(dotkeys):
         return self['index']
 
     def __enter__(self):
-        self.begin()
+        # FIXME: use a bitmask?
+        if self._mode not in ('implicit', 'explicit'):
+            raise IPDBModeError()
+        if not self._tids:
+            self.begin()
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
@@ -290,6 +352,14 @@ class interface(dotkeys):
             self.commit()
         except Exception as e:
             self.last_error = e
+            raise e
+
+    def __repr__(self):
+        res = {}
+        for i in self:
+            if self[i] is not None:
+                res[i] = self[i]
+        return res.__repr__()
 
     @property
     def if_master(self):
@@ -303,13 +373,6 @@ class interface(dotkeys):
             # vlan ports
             return self._parent.get(self['link'], None)
 
-    @property
-    def if_slaves(self):
-        '''
-        [property] Dictionary of slave interfaces
-        '''
-        return self._slaves
-
     def load(self, dev):
         '''
         Update the interface info from RTM_NEWLINK message.
@@ -318,6 +381,7 @@ class interface(dotkeys):
         changes directly into the interface data.
         '''
         self._direct_state.acquire()
+        self._exists = True
         self.update(dev)
         self._attrs = set()
         for (name, value) in dev['attrs']:
@@ -345,7 +409,7 @@ class interface(dotkeys):
         self._direct_state.release()
 
     @update
-    def __setitem__(self, key, value, direct):
+    def __setitem__(self, direct, key, value):
         if not direct:
             transaction = self.last()
             transaction[key] = value
@@ -353,7 +417,7 @@ class interface(dotkeys):
             dotkeys.__setitem__(self, key, value)
 
     @update
-    def __delitem__(self, key, direct):
+    def __delitem__(self, direct, key):
         if not direct:
             transaction = self.last()
             if key in transaction:
@@ -362,7 +426,10 @@ class interface(dotkeys):
             dotkeys.__delitem__(self, key)
 
     @update
-    def add_ip(self, ip, mask, direct):
+    def add_ip(self, direct, ip, mask=None):
+        if mask is None:
+            ip, mask = ip.split('/')
+            mask = int(mask, 0)
         if not direct:
             transaction = self.last()
             transaction.add_ip(ip, mask)
@@ -370,7 +437,10 @@ class interface(dotkeys):
             self['ipaddr'].add((ip, mask))
 
     @update
-    def del_ip(self, ip, mask, direct):
+    def del_ip(self, direct, ip, mask=None):
+        if mask is None:
+            ip, mask = ip.split('/')
+            mask = int(mask, 0)
         if not direct:
             transaction = self.last()
             if (ip, mask) in transaction['ipaddr']:
@@ -378,6 +448,27 @@ class interface(dotkeys):
         else:
             self['ipaddr'].remove((ip, mask))
 
+    @update
+    def add_port(self, direct, port):
+        if isinstance(port, interface):
+            port = port['index']
+        if not direct:
+            transaction = self.last()
+            transaction.add_port(port)
+        else:
+            self['ports'].add(port)
+
+    @update
+    def del_port(self, direct, port):
+        if isinstance(port, interface):
+            port = port['index']
+        if not direct:
+            transaction = self.last()
+            if port in transaction['ports']:
+                transaction.del_port(port)
+        else:
+            self['ports'].remove(port)
+
     def begin(self):
         '''
         Start new transaction
@@ -407,7 +498,10 @@ class interface(dotkeys):
         removed = self - self.last()
         added['-ipaddr'] = removed['ipaddr']
         added['+ipaddr'] = added['ipaddr']
+        added['-ports'] = removed['ports']
+        added['+ports'] = added['ports']
         del added['ipaddr']
+        del added['ports']
         return added
 
     def drop(self):
@@ -422,8 +516,11 @@ class interface(dotkeys):
         '''
         self._parent._addr_event.clear()
         self['ipaddr'].clear()
-        self.ip.get_links(self['index'])
-        self.ip.get_addr()
+        try:
+            self.ip.get_links(self['index'])
+            self.ip.get_addr()
+        except NetlinkQueueEmpty as e:
+            raise IPDBUnrecoverableError('lost netlink', e)
         self._parent._addr_event.wait()
 
     def commit(self, tid=None, transaction=None, rollback=False):
@@ -434,38 +531,83 @@ class interface(dotkeys):
         e = None
         added = None
         removed = None
-        snapshot = self.pick()
         if tid:
             transaction = self._transactions[tid]
         else:
             transaction = transaction or self.last()
+
+        # if the interface does not exist, create it first ;)
+        if not self._exists:
+            request = IPLinkRequest(self)
+            self._parent._links_event.clear()
+            try:
+                self.ip.link('add', **request)
+            except Exception as e:
+                # on failure, invalidate the interface and detach it
+                # from the parent
+                # 1. drop the iproute() link
+                self.ip = None
+                # 2. clean up ipdb
+                self._parent.detach(self['index'])
+                self._parent.detach(self['ifname'])
+                # 3. invalidate the interface
+                self._direct_state.acquire()
+                for i in tuple(self.keys()):
+                    del self[i]
+                self._direct_state.release()
+                # 4. the rest
+                self._mode = 'invalid'
+                # raise the exception
+                raise e
+
+            # all is OK till now, so continue
+            # we do not know what to load, so load everything
+            self.ip.get_links()
+            self._parent._links_event.wait()
+
+        # now we have our index and IP set and all other stuff
+        snapshot = self.pick()
+
         try:
-            # commit IP address changes
             removed = self - transaction
+            added = transaction - self
+
+            # 8<---------------------------------------------
+            # IP address changes
             for i in removed['ipaddr']:
                 # When you remove a primary IP addr, all subnetwork
                 # can be removed. In this case you will fail, but
                 # it is OK, no need to roll back
                 try:
-                    self.ip.addr_del(self['index'], i[0], i[1])
-                except netlink_error as x:
+                    self.ip.addr('delete', self['index'], i[0], i[1])
+                except NetlinkSocketError as x:
                     # bypass only errno 99, 'Cannot assign address'
                     if x.errno != 99:
                         raise x
 
-            added = transaction - self
             for i in added['ipaddr']:
-                self.ip.addr_add(self['index'], i[0], i[1])
+                self.ip.addr('add', self['index'], i[0], i[1])
+            # 8<---------------------------------------------
+            # Interface slaves
+            for i in removed['ports']:
+                # detach the port
+                self.ip.link('set', index=i, master=0)
 
-            # commit interface changes
-            request = {}
+            for i in added['ports']:
+                # enslave the port
+                self.ip.link('set', index=i, master=self['index'])
+
+            # 8<---------------------------------------------
+            # Interface changes
+            request = IPLinkRequest()
             for key in added:
                 if key in nla_fields:
                     request[key] = added[key]
 
             # apply changes only if there is something to apply
             if request:
-                self.ip.link('set', self['index'], **request)
+                self.ip.link('set', index=self['index'], **request)
+            # 8<---------------------------------------------
 
         except Exception as e:
             # something went wrong: roll the transaction back
@@ -569,6 +711,7 @@ class ipdb(dotkeys):
 
         # update events
         self._addr_event = threading.Event()
+        self._links_event = threading.Event()
 
         # load information on startup
         links = self.ip.get_links()
@@ -580,7 +723,6 @@ class ipdb(dotkeys):
         self.ip.monitor()
         self.ip.mirror()
         self._mthread = threading.Thread(target=self.monitor)
-        self._mthread.setDaemon(True)
         self._mthread.start()
 
     def __dir__(self):
@@ -592,7 +734,42 @@ class ipdb(dotkeys):
     def shutdown(self):
         self._stop = True
         self.ip.get_links()
-        self.ip.shutdown_sockets()
+        self.ip.shutdown()
+        self._mthread.join()
+
+    def create(self, kind, ifname, **kwarg):
+        '''
+        Create an interface. Arguments 'kind' and 'ifname' are
+        required.
+
+        * kind -- interface type, can be of:
+          * bridge
+          * bond
+          * vlan
+          * tun
+          * dummy
+        * ifname -- interface name
+
+        Different interface kinds can require different
+        arguments for creation.
+
+        FIXME: this should be documented.
+        '''
+        i = interface(ipr=self.ip, parent=self, mode='snapshot')
+        i['kind'] = kind
+        i['index'] = kwarg.get('index', 0)
+        i['ifname'] = ifname
+        self.by_name[i['ifname']] = self[i['ifname']] = i
+        i.update(kwarg)
+        i._mode = self.mode
+        i.begin()
+        return i
+
+    def detach(self, item):
+        if item in self:
+            del self[item]
+            if item in self.ipaddr:
+                del self.ipaddr[item]
 
     def update_links(self, links):
         '''
@@ -600,11 +777,13 @@ class ipdb(dotkeys):
         '''
         for dev in links:
             if dev['index'] not in self.ipaddr:
-                self.ipaddr[dev['index']] = ipset()
+                self.ipaddr[dev['index']] = LinkedSet()
             i = \
                 self.by_index[dev['index']] = \
                 self[dev['index']] = \
-                interface(dev, self.ip, self, mode=self.mode)
+                self.get(dev.get_attr('IFLA_IFNAME')[0], None) or \
+                interface(ipr=self.ip, parent=self, mode=self.mode)
+            i.load(dev)
             self[i['ifname']] = \
                 self.by_name[i['ifname']] = i
             self.old_names[dev['index']] = i['ifname']
@@ -619,17 +798,24 @@ class ipdb(dotkeys):
             if master:
                 master = self[master[0]]
                 index = msg['index']
-                name = msg.get_attr('IFLA_IFNAME')[0]
                 if msg['event'] == 'RTM_NEWLINK':
-                    master._slaves[index] = self[msg['index']]
-                    master._slaves[name] = self[msg['index']]
+                    # TODO tags: ipdb
+                    # The code serves one particular case, when
+                    # an enslaved interface is set to belong to
+                    # another master. In this case there will be
+                    # no 'RTM_DELLINK', only 'RTM_NEWLINK', and
+                    # we can end up in a broken state, when two
+                    # masters refers to the same slave
+                    for device in self.by_index:
+                        if index in self[device]['ports']:
+                            self[device]['ports'].remove(index)
+                    # FIXME: move this to a dedicated method
+                    master['ports'].add(index)
                 elif msg['event'] == 'RTM_DELLINK':
-                    if index in master._slaves:
-                        del master._slaves[index]
-                    if name in master._slaves:
-                        del master._slaves[name]
-                    if 'master' in self[msg['index']]:
-                        del self[msg['index']]['master']
+                    # TODO tags: ipdb
+                    # FIXME: move this to a dedicated method
+                    if index in master['ports']:
+                        master['ports'].remove(index)
 
     def update_addr(self, addrs, action='add'):
         '''
@@ -668,8 +854,6 @@ class ipdb(dotkeys):
                         if self[index]['ifname'] != old:
                             # FIXME catch exception
                             # FIXME isolate dict updates
-                            if self[old].if_master:
-                                del self[old].if_master.if_slaves[old]
                             del self[old]
                             del self.by_name[old]
                             if index in self.old_names:
@@ -680,6 +864,8 @@ class ipdb(dotkeys):
                     else:
                         self.update_links([msg])
                     self.update_slaves([msg])
+                    # what about removal?
+                    self._links_event.set()
                 elif msg.get('event', None) == 'RTM_DELLINK':
                     self.update_slaves([msg])
                     if msg['change'] == 0xffffffff:
